//
// Created by vlad on 18.02.2021.
//

#include <WiFi.h>
#include "web.h"

#include "config.h"

#include <http/HTTPSServer.hpp>
#include <string>

using namespace httpsserver;


static void getHTML(HTTPRequest * req, HTTPResponse * res) {
    static char html[] = "<!DOCTYPE html>\n"
                         "<html lang=\"ru\">\n"
                         "<head>\n"
                         "    <meta charset=\"UTF-8\">\n"
                         "    <meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0\">\n"
                         "    <title>Esp32 FM pirate</title>\n"
                         "    <style>\n"
                         "        * {\n"
                         "            color: #7295c4;\n"
                         "            border-color: #353535;\n"
                         "            background: #121212;\n"
                         "        }\n"
                         "        h1 {\n"
                         "            font-size: 18px;\n"
                         "        }\n"
                         "        .main {\n"
                         "            background: #212121;\n"
                         "            height: auto;\n"
                         "        }\n"
                         "        .main header {\n"
                         "            height: auto;\n"
                         "            padding: 15px 16px;\n"
                         "            justify-content: center;\n"
                         "        }\n"
                         "        .main .content {\n"
                         "            padding: 20px;\n"
                         "        }\n"
                         "        .main {\n"
                         "            border: 3px solid;\n"
                         "            border-radius: 8px;\n"
                         "            overflow: hidden;\n"
                         "            width: 22em;\n"
                         "            min-width: 300px;\n"
                         "            margin: 50px auto;\n"
                         "            height: auto;\n"
                         "            text-align: center;\n"
                         "        }\n"
                         "        .form-row {\n"
                         "            padding: 4px 0;\n"
                         "        }\n"
                         "        .form-row > label {\n"
                         "            display: block;\n"
                         "            line-height: 2em;\n"
                         "        }\n"
                         "        .form-row > input {\n"
                         "            padding: 8px;\n"
                         "            width: 100%;\n"
                         "            box-sizing: border-box;\n"
                         "        }\n"
                         "        input {\n"
                         "            border: 2px solid;\n"
                         "            border-radius: 4px;\n"
                         "        }\n"
                         "\n"
                         "        /* далее стили для слидеров */\n"
                         "        .switch {\n"
                         "            position: relative;\n"
                         "            display: inline-block;\n"
                         "            width: 60px;\n"
                         "            height: 34px;\n"
                         "            margin: 0 auto;\n"
                         "        }\n"
                         "\n"
                         "        .switch input {\n"
                         "            opacity: 0;\n"
                         "            width: 0;\n"
                         "            height: 0;\n"
                         "        }\n"
                         "\n"
                         "        .slider {\n"
                         "            position: absolute;\n"
                         "            cursor: pointer;\n"
                         "            top: 0;\n"
                         "            left: 0;\n"
                         "            right: 0;\n"
                         "            bottom: 0;\n"
                         "            background-color: #ccc;\n"
                         "            -webkit-transition: .4s;\n"
                         "            transition: .4s;\n"
                         "        }\n"
                         "\n"
                         "        .slider:before {\n"
                         "            position: absolute;\n"
                         "            content: \"\";\n"
                         "            height: 26px;\n"
                         "            width: 26px;\n"
                         "            left: 4px;\n"
                         "            bottom: 4px;\n"
                         "            background-color: white;\n"
                         "            -webkit-transition: .4s;\n"
                         "            transition: .4s;\n"
                         "        }\n"
                         "\n"
                         "        input:checked + .slider {\n"
                         "            background-color: #2196F3;\n"
                         "        }\n"
                         "\n"
                         "        input:checked + .slider:before {\n"
                         "            transform: translateX(26px);\n"
                         "        }\n"
                         "    </style>\n"
                         "</head>\n"
                         "<body>\n"
                         "<div class=\"main\">\n"
                         "    <header><h1>Esp32 FM pirate</h1></header>\n"
                         "    <div class=\"content\">\n"
                         "        <div id=\"form-content\">\n"
                         "\n"
                         "            <div class=\"form-row\">\n"
                         "                <label class=\"required\" for=\"control-rds\">RDS текст</label>\n"
                         "                <input type=\"text\" maxlength=\"16\" size=\"20\" id=\"control-rds\" onchange=\"sendParam('rds', document.getElementById('control-rds').value)\">\n"
                         "            </div>\n"
                         "\n"
                         "            <div class=\"form-row\">\n"
                         "                <label class=\"required\" for=\"control-frq\">Частота (96.00fm = 9600)</label>\n"
                         "                <input type=\"number\" maxlength=\"5\" id=\"control-frq\" onchange=\"sendParam('frq', document.getElementById('control-frq').value)\">\n"
                         "            </div>\n"
                         "\n"
                         "            <div class=\"form-row\">\n"
                         "                <label for=\"control-fmen\">FM Enable</label>\n"
                         "                <label class=\"switch\">\n"
                         "                    <input type=\"checkbox\" id=\"control-fmen\" onchange=\"sendParam('fmen', document.getElementById('control-fmen').checked ? 1 : 0)\">\n"
                         "                    <span class=\"slider\"></span>\n"
                         "                </label>\n"
                         "            </div>\n"
                         "\n"
                         "            <script>\n"
                         "                function sendParam(io_interface, value) {\n"
                         "                    let xhr = new XMLHttpRequest();\n"
                         "                    let url = `/`\n"
                         "                    if (io_interface !== null) {\n"
                         "                        url += `?${io_interface}=${encodeURIComponent(value)}`\n"
                         "                    }\n"
                         "                    xhr.onload = function () {\n"
                         "                        if (xhr.status !== 200) {\n"
                         "                            alert(`Ошибка ${xhr.status}: ${xhr.statusText}`);\n"
                         "                        } else {\n"
                         "                            let res = JSON.parse(xhr.response)\n"
                         "                            if (\"frq\" in res) {\n"
                         "                                document.getElementById('control-frq').value = res[\"frq\"]\n"
                         "                            }\n"
                         "                            if (\"rds\" in res) {\n"
                         "                                document.getElementById('control-rds').value = res[\"rds\"]\n"
                         "                            }\n"
                         "                            if (\"fmen\" in res) {\n"
                         "                                document.getElementById('control-fmen').checked = !!res[\"fmen\"]\n"
                         "                            }\n"
                         "                        }\n"
                         "                    }\n"
                         "                    xhr.open('POST', url)\n"
                         "                    xhr.send()\n"
                         "                }\n"
                         "\n"
                         "                sendParam(null)\n"
                         "            </script>\n"
                         "        </div>\n"
                         "    </div>\n"
                         "</div>\n"
                         "</body>\n"
                         "</html>";
    // ответим
    req->discardRequestBody();
    res->setStatusCode(200);
    res->setHeader("Content-Type", "text/html; charset=utf-8");
    res->setHeader("Content-Length", intToString(sizeof(html) - 1));
    // отправим страничку

//    res->printf(html, pwm_values[0], pwm_values[1], pwm_values[2], pwm_values[3]);
    res->write((const uint8_t*)html, sizeof(html) - 1);
}

static void getFavicon(HTTPRequest * req, HTTPResponse * res) {
    static const uint8_t favicon_data[] = {
            0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A, 0x00, 0x00, 0x00, 0x0D, 0x49, 0x48, 0x44, 0x52,
            0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x08, 0x02, 0x00, 0x00, 0x00, 0xFC, 0x18, 0xED,
            0xA3, 0x00, 0x00, 0x01, 0x84, 0x69, 0x43, 0x43, 0x50, 0x49, 0x43, 0x43, 0x20, 0x70, 0x72, 0x6F,
            0x66, 0x69, 0x6C, 0x65, 0x00, 0x00, 0x28, 0x91, 0x7D, 0x91, 0x3D, 0x48, 0xC3, 0x40, 0x18, 0x86,
            0xDF, 0xA6, 0x4A, 0x8B, 0x54, 0x1C, 0xEC, 0x20, 0x22, 0x92, 0xA1, 0x3A, 0x59, 0x10, 0x15, 0x71,
            0xD4, 0x2A, 0x14, 0xA1, 0x42, 0xA8, 0x15, 0x5A, 0x75, 0x30, 0xB9, 0xF4, 0x0F, 0x9A, 0x34, 0x24,
            0x29, 0x2E, 0x8E, 0x82, 0x6B, 0xC1, 0xC1, 0x9F, 0xC5, 0xAA, 0x83, 0x8B, 0xB3, 0xAE, 0x0E, 0xAE,
            0x82, 0x20, 0xF8, 0x03, 0xE2, 0xE8, 0xE4, 0xA4, 0xE8, 0x22, 0x25, 0x7E, 0x97, 0x14, 0x5A, 0xC4,
            0x78, 0xC7, 0x71, 0x0F, 0xEF, 0x7D, 0xEF, 0xCB, 0xDD, 0x77, 0x80, 0xD0, 0xA8, 0x30, 0xCD, 0xEA,
            0x1A, 0x07, 0x34, 0xDD, 0x36, 0xD3, 0xC9, 0x84, 0x98, 0xCD, 0xAD, 0x8A, 0xA1, 0x57, 0x04, 0x11,
            0x42, 0x98, 0xE6, 0xB0, 0xCC, 0x2C, 0x63, 0x4E, 0x92, 0x52, 0xF0, 0x1D, 0x5F, 0xF7, 0x08, 0xF0,
            0xFD, 0x2E, 0xCE, 0xB3, 0xFC, 0xEB, 0xFE, 0x1C, 0xBD, 0x6A, 0xDE, 0x62, 0x40, 0x40, 0x24, 0x9E,
            0x65, 0x86, 0x69, 0x13, 0x6F, 0x10, 0x4F, 0x6F, 0xDA, 0x06, 0xE7, 0x7D, 0xE2, 0x28, 0x2B, 0xC9,
            0x2A, 0xF1, 0x39, 0xF1, 0x98, 0x49, 0x17, 0x24, 0x7E, 0xE4, 0xBA, 0xE2, 0xF1, 0x1B, 0xE7, 0xA2,
            0xCB, 0x02, 0xCF, 0x8C, 0x9A, 0x99, 0xF4, 0x3C, 0x71, 0x94, 0x58, 0x2C, 0x76, 0xB0, 0xD2, 0xC1,
            0xAC, 0x64, 0x6A, 0xC4, 0x53, 0xC4, 0x31, 0x55, 0xD3, 0x29, 0x5F, 0xC8, 0x7A, 0xAC, 0x72, 0xDE,
            0xE2, 0xAC, 0x55, 0x6A, 0xAC, 0x75, 0x4F, 0xFE, 0xC2, 0x48, 0x5E, 0x5F, 0x59, 0xE6, 0x3A, 0xAD,
            0x21, 0x24, 0xB1, 0x88, 0x25, 0x48, 0x10, 0xA1, 0xA0, 0x86, 0x32, 0x2A, 0xB0, 0x11, 0xA7, 0x5D,
            0x27, 0xC5, 0x42, 0x9A, 0xCE, 0x13, 0x3E, 0xFE, 0x41, 0xD7, 0x2F, 0x91, 0x4B, 0x21, 0x57, 0x19,
            0x8C, 0x1C, 0x0B, 0xA8, 0x42, 0x83, 0xEC, 0xFA, 0xC1, 0xFF, 0xE0, 0x77, 0x6F, 0xAD, 0xC2, 0xE4,
            0x84, 0x97, 0x14, 0x49, 0x00, 0xDD, 0x2F, 0x8E, 0xF3, 0x31, 0x02, 0x84, 0x76, 0x81, 0x66, 0xDD,
            0x71, 0xBE, 0x8F, 0x1D, 0xA7, 0x79, 0x02, 0x04, 0x9F, 0x81, 0x2B, 0xBD, 0xED, 0xAF, 0x36, 0x80,
            0x99, 0x4F, 0xD2, 0xEB, 0x6D, 0x2D, 0x76, 0x04, 0xF4, 0x6D, 0x03, 0x17, 0xD7, 0x6D, 0x4D, 0xD9,
            0x03, 0x2E, 0x77, 0x80, 0x81, 0x27, 0x43, 0x36, 0x65, 0x57, 0x0A, 0xD2, 0x12, 0x0A, 0x05, 0xE0,
            0xFD, 0x8C, 0xBE, 0x29, 0x07, 0xF4, 0xDF, 0x02, 0x3D, 0x6B, 0x5E, 0xDF, 0x5A, 0xE7, 0x38, 0x7D,
            0x00, 0x32, 0xD4, 0xAB, 0xD4, 0x0D, 0x70, 0x70, 0x08, 0x8C, 0x16, 0x29, 0x7B, 0xDD, 0xE7, 0xDD,
            0xE1, 0xCE, 0xBE, 0xFD, 0x5B, 0xD3, 0xEA, 0xDF, 0x0F, 0xFA, 0xE8, 0x72, 0x77, 0xE0, 0x23, 0xEF,
            0x60, 0x00, 0x00, 0x00, 0x09, 0x70, 0x48, 0x59, 0x73, 0x00, 0x00, 0x2E, 0x23, 0x00, 0x00, 0x2E,
            0x23, 0x01, 0x78, 0xA5, 0x3F, 0x76, 0x00, 0x00, 0x00, 0x07, 0x74, 0x49, 0x4D, 0x45, 0x07, 0xE6,
            0x03, 0x07, 0x08, 0x0C, 0x1F, 0xF9, 0x10, 0x03, 0x17, 0x00, 0x00, 0x00, 0x19, 0x74, 0x45, 0x58,
            0x74, 0x43, 0x6F, 0x6D, 0x6D, 0x65, 0x6E, 0x74, 0x00, 0x43, 0x72, 0x65, 0x61, 0x74, 0x65, 0x64,
            0x20, 0x77, 0x69, 0x74, 0x68, 0x20, 0x47, 0x49, 0x4D, 0x50, 0x57, 0x81, 0x0E, 0x17, 0x00, 0x00,
            0x00, 0x96, 0x49, 0x44, 0x41, 0x54, 0x48, 0xC7, 0xED, 0x55, 0x51, 0x0E, 0x80, 0x20, 0x08, 0x85,
            0xD6, 0x09, 0xCC, 0xEB, 0xD8, 0x95, 0xBB, 0x43, 0x5D, 0xA7, 0x3A, 0x82, 0xF4, 0xE9, 0x40, 0xCD,
            0x5C, 0x6B, 0xD3, 0x06, 0x5F, 0x6C, 0x30, 0x71, 0x3C, 0xDE, 0x7B, 0x48, 0xF0, 0x6D, 0x0C, 0xD0,
            0xFB, 0x80, 0x31, 0xA4, 0x3B, 0x2B, 0xA0, 0x95, 0xAD, 0xC4, 0x3B, 0x90, 0xD8, 0xE7, 0xE8, 0x30,
            0xAC, 0xDB, 0xFE, 0x66, 0x45, 0xFD, 0x0F, 0x40, 0x0A, 0xD9, 0xC9, 0x2B, 0x26, 0x02, 0xD9, 0xB1,
            0x3A, 0x6C, 0x37, 0xEF, 0x92, 0x62, 0xD0, 0x22, 0x93, 0x09, 0x3D, 0x2F, 0x39, 0x8E, 0xDA, 0x1A,
            0xDD, 0x07, 0x56, 0x4A, 0x45, 0x65, 0x88, 0xA3, 0x02, 0x58, 0x20, 0xA5, 0x15, 0x0A, 0x72, 0x05,
            0x06, 0x08, 0xF6, 0x09, 0x33, 0x5F, 0xF8, 0x41, 0x41, 0x53, 0xE6, 0x82, 0x4E, 0x08, 0x7B, 0x40,
            0x65, 0x72, 0x8B, 0x52, 0x21, 0x4C, 0x7F, 0x12, 0xF6, 0xE0, 0x93, 0x44, 0xCD, 0x9A, 0xBE, 0x62,
            0xD0, 0xA2, 0xE9, 0xEB, 0x8A, 0x32, 0x71, 0x01, 0xBC, 0x3B, 0x1E, 0x5A, 0xC9, 0x99, 0x48, 0x22,
            0x00, 0x00, 0x00, 0x00, 0x49, 0x45, 0x4E, 0x44, 0xAE, 0x42, 0x60, 0x82,
    };
    // ответим
    req->discardRequestBody();
    res->setStatusCode(200);
    res->setHeader("Content-Type", "image/png");
    res->setHeader("Content-Length", intToString(sizeof(favicon_data)));
    // отправим иконку
    res->write(favicon_data, sizeof(favicon_data));
}

uint32_t output_frq = 9600;
std::string output_rds = "FUCK YOU!1!";
bool output_enable = false;

static void setOutputData(HTTPRequest * req, HTTPResponse * res) {
    auto p = req->getParams();
    if (p != nullptr) {
        // сначала частота
        std::string param;
        p->getQueryParameter("frq", param);
        if (!param.empty()) {
            auto val = parseInt(param);
            val = min(10800, max(8750, val));
            output_frq = val;
            radio->tuneFM(val);
            // TODO прикрутить запись частоты в передатчик
            IF_DEBUG Serial.printf("[web] _set_frq: set frq=%d\n", val);
        }

        // теперь RDS
        param.clear();
        p->getQueryParameter("rds", param);
        if (!param.empty()) {
            output_rds = param;
            // TODO прикрутить запись rds в передатчик
            IF_DEBUG Serial.printf("[web] _set_rds: set rds=\"%s\"\n", param.c_str());
        }

        // теперь включен/выключен
        param.clear();
        p->getQueryParameter("fmen", param);
        if (!param.empty()) {
            bool val = parseInt(param) != 0;
            output_enable = val;
            // TODO прикрутить запись флага в передатчик
            IF_DEBUG Serial.printf("[web] _set_fmen: set fmen=%d\n", val);
        }
    } else {
        IF_DEBUG Serial.printf("[web] setOutputData: empty params!\n");
    }

    // сделаем json ответ со всеми интерфейсами
    static char response_buffer[1024];
    sprintf(response_buffer, R"({"frq": %d, "rds": "%s", "fmen": %d})", output_frq, output_rds.c_str(), output_enable);

    // отправим json ответ
    res->setStatusCode(200);
    res->setHeader("Content-Type", "application/json; charset=utf-8");
    auto len = (uint32_t)strlen(response_buffer);
    res->setHeader("Content-Length", intToString((int32_t)len));

    res->write((const uint8_t*)response_buffer, len);

    IF_DEBUG Serial.printf("[web] send json: %s\n", response_buffer);
}


static void handle404(HTTPRequest * req, HTTPResponse * res) {
    // Discard request body, if we received any
    // We do this, as this is the default node and may also server POST/PUT requests
    req->discardRequestBody();
    IF_DEBUG Serial.printf("[handle404] Resource '%s' not found\n", req->getRequestString().c_str());

    // Set the response status
    res->setStatusCode(404);
    res->setStatusText("Not Found");

    // Set content type of the response
    res->setHeader("Content-Type", "text/html");

    // Write a tiny HTML page
    res->printf("<!DOCTYPE html>\n"
                 "<html data-theme=\"dark\" lang=\"ru\"><head><meta charset=\"UTF-8\">"
                 "<link rel=\"stylesheet\" href=\"style.css\">"
                 "<title>Not Found</title></head><body><h1>404 Not Found</h1>"
                 "<p>Вы запросили несуществующую страницу...</p></body></html>");
}

[[noreturn]] void webThread() {
    IF_DEBUG {
        Serial.printf("******** START SERVER ********\n");
    }

//    WiFiClass::mode(WIFI_MODE_STA);
    WiFi.softAP(WIFI_SSID, WIFI_PSK);
//    WiFi.begin(WIFI_SSID, WIFI_PSK);
    Serial.printf("[webThread] IP address: %s\n", WiFi.softAPIP().toString().c_str());
//    Serial.printf("[webThread] IP address: %s\n", WiFi.localIP().toString().c_str());

    HTTPServer server = HTTPServer();
    // Add all nodes to the server so they become accessible:
    server.registerNode(new ResourceNode("/", "GET", &getHTML));
    server.registerNode(new ResourceNode("/favicon.ico", "GET", &getFavicon));
    server.registerNode(new ResourceNode("/", "POST", &setOutputData));
    server.setDefaultNode(new ResourceNode("", "GET", &handle404));

    server.start();
    Serial.printf("[webThread] Server is running\n");

    for (;;) {
        // This call will let the server do its work
        server.loop();
        // Other code would go here...
        vTaskDelay(1);
    }
}

